#include "stdafx.h"
#include "shlwapi.h"
#include "CAttachmentDropTargetCtrl.hpp"
#include "COleDropWndTarget.hpp"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

//
// Constructor.
//
COleDropWndTarget::COleDropWndTarget()
{
	CF_FILECONTENTS =
		(CLIPFORMAT)RegisterClipboardFormat(CFSTR_FILECONTENTS)
	;
	CF_FILEGROUPDESCRIPTOR =
		(CLIPFORMAT)RegisterClipboardFormat(CFSTR_FILEDESCRIPTORA)
	;
}

//
// Destructor.
//
COleDropWndTarget::~COleDropWndTarget()
{
	//
}

//
// Called when a drag cursor enters a window.
//
DROPEFFECT COleDropWndTarget::OnDragEnter(
	CWnd *pWnd,
	COleDataObject *pDataObject,
	DWORD dwKeyState,
	CPoint point
)
{
	return IsFileDrop(pDataObject) ? DROPEFFECT_COPY : DROPEFFECT_NONE;
}

//
// Called when the cursor leaves a window.
//
void COleDropWndTarget::OnDragLeave(CWnd *pWnd)
{

	// Call base implementation.
	COleDropTarget::OnDragLeave(pWnd);
}

//
// Called when the cursor is dragged over a window.
//
DROPEFFECT COleDropWndTarget::OnDragOver(
	CWnd *pWnd,
	COleDataObject *pDataObject,
	DWORD dwKeyState,
	CPoint point
)
{
	return IsFileDrop(pDataObject) ? DROPEFFECT_COPY : DROPEFFECT_NONE;
}

//
// Called when an item is dropped in a window.
//
BOOL COleDropWndTarget::OnDrop(
	CWnd *pWnd,
	COleDataObject *pDataObject,
	DROPEFFECT dropEffect,
	CPoint point
)
{

	// Attachments from Outlook Express or normal files.
	if (pDataObject->IsDataAvailable(CF_HDROP))
		return ProcessHdrop(pWnd, pDataObject);

	// Attachments from Outlook.
	if (
		pDataObject->IsDataAvailable(CF_FILEGROUPDESCRIPTOR) &&
		pDataObject->IsDataAvailable(CF_FILECONTENTS)
	)
		return ProcessFileGroup(pWnd, pDataObject);
	return FALSE;
}

//
// Stores the contents of a byte array into a variant containing a safe
// array of variants of type 'VT_UI1'.
//
BOOL ByteArrayToVariant(BYTE *pbSource, LONG cBytes, VARIANT *pvResult)
{
	HRESULT hr;
	BYTE HUGEP *pByte = pbSource;

	VariantInit(pvResult);

	// Create safe array of 'VARIANT' elements (indices 0 to <cBytes> - 1).
	SAFEARRAY *psa = SafeArrayCreateVector(VT_VARIANT, 0, cBytes);
	if (psa == NULL)
		return FALSE;

	// Store elements as 'VARIANT' in the safe array.
	for (LONG i = 0; i < cBytes; i++)
	{
		VARIANT *pv = new VARIANT;
		VariantInit(pv);
		pv->vt = VT_UI1;
		pv->bVal = *pByte;
		hr = SafeArrayPutElement(psa, &i, pv);
		VariantClear(pv);
		delete pv;
		if (FAILED(hr))
			return FALSE;
		pByte++;
	}

	// Get a pointer to the elements of the array.
	hr = SafeArrayAccessData(psa, (void HUGEP **)&pByte);
	if (FAILED(hr))
		return FALSE;

	pvResult->vt = VT_ARRAY | VT_VARIANT;
	pvResult->parray = psa;
	return TRUE;
}

//
// Checks whether or not a data object contains files which can be added as
// attachments.
//
BOOL COleDropWndTarget::IsFileDrop(COleDataObject *pDataObject)
{
	BOOL fSuccess = TRUE;
	if (
		pDataObject->IsDataAvailable(CF_FILEGROUPDESCRIPTOR) &&
		pDataObject->IsDataAvailable(CF_FILECONTENTS)
	)
		;
	else if (pDataObject->IsDataAvailable(CF_HDROP))
	{

		// Try to open the files.  If at least one file can be opened, the drop
		// is successful.
		FORMATETC fmt =
			{
				CF_HDROP,
				NULL,
				DVASPECT_CONTENT,
				-1,
				TYMED_HGLOBAL
			}
		;
		STGMEDIUM stgm;
		if (!pDataObject->GetData(CF_HDROP, &stgm, &fmt))
			fSuccess = FALSE;
		else
		{
			HDROP hDrop = (HDROP)stgm.hGlobal;
			UINT cFileCount = DragQueryFile(hDrop , 0xFFFFFFFF, NULL, 0);
			if (cFileCount == 0)
				fSuccess = FALSE;
			else
			{
				BOOL fValidFile = FALSE;
				for (UINT i = 0; i < cFileCount; i++)
				{
					TCHAR lpszPath[MAX_PATH];
					UINT n = DragQueryFile(hDrop, i, lpszPath, sizeof(lpszPath));
					HANDLE hFile =
						CreateFile(
							lpszPath,
							GENERIC_READ,
							FILE_SHARE_READ | FILE_SHARE_WRITE,
							NULL,
							OPEN_EXISTING,
							FILE_ATTRIBUTE_NORMAL,
							NULL
						)
					;
					if (hFile != INVALID_HANDLE_VALUE)
					{
						fValidFile = TRUE;
						CloseHandle(hFile);
						break;
					}
				}
				if (!fValidFile)
					fSuccess = FALSE;
			}
		}
		ReleaseStgMedium(&stgm);
	}
	else
		fSuccess = FALSE;
	return fSuccess;
}

//
// Processes a group of files stored as 'CF_HDROP' and raises events to
// pass data to the client using the control.
//
BOOL COleDropWndTarget::ProcessHdrop(
	CWnd *pWnd,
	COleDataObject *pDataObject
)
{
	BOOL fSuccess = TRUE;
	FORMATETC fmt =
		{
			CF_HDROP,
			NULL,
			DVASPECT_CONTENT,
			-1,
			TYMED_HGLOBAL
		}
	;
	STGMEDIUM stgm;
	if (!pDataObject->GetData(CF_HDROP, &stgm, &fmt))
		fSuccess = FALSE;
	else
	{
		HDROP hDrop = (HDROP)stgm.hGlobal;
		UINT cFileCount = DragQueryFile(hDrop , 0xFFFFFFFF, NULL, 0);
		if (cFileCount == 0)
			fSuccess = FALSE;
		else
		{
			for (UINT i = 0; i < cFileCount; i++)
			{
				TCHAR lpszPath[MAX_PATH];
				UINT n = DragQueryFile(hDrop, i, lpszPath, sizeof(lpszPath));

				// Determine file contents.
				HANDLE hFile =
					CreateFile(
						lpszPath,
						GENERIC_READ,
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						NULL,
						OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL,
						NULL
					)
				;
				if (hFile != INVALID_HANDLE_VALUE)
				{
					LPTSTR lpszFileName = PathFindFileName(lpszPath);
					((CAttachmentDropTargetCtrl *)pWnd)->FileAdded(lpszFileName);
					DWORD cdwBlockSize = ((CAttachmentDropTargetCtrl *)pWnd)->GetBlockSize();
					BYTE *pbData = (BYTE *)malloc(cdwBlockSize);
					if (pbData == NULL)
						fSuccess = FALSE;
					else
					{
						DWORD cdwBytesRead;
						do
						{
							if (!ReadFile(hFile, pbData, cdwBlockSize, &cdwBytesRead, NULL))
								fSuccess = FALSE;
							else
							{
								VARIANT *pv = new VARIANT;
								if (ByteArrayToVariant(pbData, cdwBytesRead, pv))
								{
									((CAttachmentDropTargetCtrl *)pWnd)->DataArrived(pv);
									SafeArrayDestroy(pv->parray);
								}
								else
									fSuccess = FALSE;
								VariantClear(pv);
								delete pv;
							}
						}
						while (cdwBytesRead == cdwBlockSize && fSuccess);
						free(pbData);
					}
					CloseHandle(hFile);
				}
			}
		}
		ReleaseStgMedium(&stgm);
	}
	return fSuccess;
}

//
// Processes a group of files stored as 'CF_FILEGROUPDESCRIPTOR' and
// 'CF_FILECONTENTS' and raises events to pass data to the client using the
// control.
//
BOOL COleDropWndTarget::ProcessFileGroup(
	CWnd *pWnd,
	COleDataObject *pDataObject
)
{
	BOOL fSuccess = TRUE;

	// Determine file group descriptor.
	FORMATETC fmt =
		{
			CF_FILEGROUPDESCRIPTOR,
			NULL,
			DVASPECT_CONTENT,
			-1,
			TYMED_HGLOBAL
		}
	;
	STGMEDIUM stgm;
	if (!pDataObject->GetData(CF_FILEGROUPDESCRIPTOR, &stgm, &fmt))
		fSuccess = FALSE;
	else
	{
		HGLOBAL hgData = GlobalLock(stgm.hGlobal);
		LPFILEGROUPDESCRIPTORA lpfgd = (LPFILEGROUPDESCRIPTORA)hgData;
		for (UINT i = 0; i < lpfgd->cItems; i++)
		{
			LPFILEDESCRIPTORA lpfd = &lpfgd->fgd[i];

			// TODO: 'cFileName' is ANSI on Windows XP, but maybe Unicode on
			// Windows Server 2003!
			((CAttachmentDropTargetCtrl *)pWnd)->FileAdded(CA2CT(lpfd->cFileName));

			// Determine file contents.
			FORMATETC fmt =
				{
					CF_FILECONTENTS,
					NULL,
					DVASPECT_CONTENT,
					i,
					TYMED_HGLOBAL | TYMED_ISTORAGE | TYMED_ISTREAM
				}
			;
			STGMEDIUM stgm;
			if (!pDataObject->GetData(CF_FILECONTENTS, &stgm, &fmt))
				fSuccess = FALSE;
			else
			{
				switch (stgm.tymed)
				{
				case TYMED_HGLOBAL:
					fSuccess = FALSE;
					break;
				case TYMED_ISTORAGE:
					fSuccess = FALSE;
					break;
				case TYMED_ISTREAM:
					IStream *pStream = stgm.pstm;
					ULONG cBlockSize =
						(ULONG)((CAttachmentDropTargetCtrl *)pWnd)->GetBlockSize()
					;
					BYTE *pbData = (BYTE *)malloc(cBlockSize);
					if (pbData != NULL)
					{
						ULONG cBytesRead;
						do
						{
							if (pStream->Read(pbData, cBlockSize, &cBytesRead) == S_OK)
							{
								VARIANT *pv = new VARIANT;
								if (ByteArrayToVariant(pbData, cBytesRead, pv))
								{
									((CAttachmentDropTargetCtrl *)pWnd)->DataArrived(pv);
									SafeArrayDestroy(pv->parray);
								}
								else
									fSuccess = FALSE;
								VariantClear(pv);
								delete pv;
							}
						}
						while (cBytesRead == cBlockSize && fSuccess);
						free(pbData);
					}
					break;
				}
				ReleaseStgMedium(&stgm);
			}
		}
		GlobalUnlock(hgData);
		ReleaseStgMedium(&stgm);
	}
	return fSuccess;
}