#include "stdafx.h"
#include "comcat.h"

//
// Helper function to create a component category and associated
// description.
//
HRESULT CreateComponentCategory(CATID catid, WCHAR* lpszCategoryDescription)
{
	ICatRegister *pcr = NULL;
	HRESULT hr = S_OK;

	hr =
		CoCreateInstance(
			CLSID_StdComponentCategoriesMgr,
			NULL,
			CLSCTX_INPROC_SERVER,
			IID_ICatRegister,
			(void **)&pcr
		)
	;
	if (FAILED(hr))
		return hr;

	// Make sure the 'HKCR\Component Categories\{<catid>}' key is registered.
	CATEGORYINFO catinfo;
	catinfo.catid = catid;
	catinfo.lcid = 0x0409;	// English.

	// Make sure the provided description is not too long.  Only copy the
	// first 127 characters if it is.
	size_t len = wcslen(lpszCategoryDescription);
	if (len > 127)
		len = 127;
	wcsncpy_s(catinfo.szDescription, lpszCategoryDescription, len);

	// Make sure the description is null terminated.
	catinfo.szDescription[len] = '\0';

	hr = pcr->RegisterCategories(1, &catinfo);
	pcr->Release();
	
	return hr;
}

//
// Helper function to register a CLSID as belonging to a component
// category.
//
HRESULT RegisterCLSIDInCategory(REFCLSID clsid, CATID catid)
{

	// Register your component categories information.
	ICatRegister *pcr = NULL;
	HRESULT hr = S_OK;
	hr =
		CoCreateInstance(
			CLSID_StdComponentCategoriesMgr,
			NULL,
			CLSCTX_INPROC_SERVER,
			IID_ICatRegister,
			(void **)&pcr
		)
	;
	if (SUCCEEDED(hr))
	{

		// Register this category as being "implemented" by the class.
		CATID rgcatid[1];
		rgcatid[0] = catid;
		hr = pcr->RegisterClassImplCategories(clsid, 1, rgcatid);
	}
	if (pcr != NULL)
		pcr->Release();
	return hr;
}